@@ -43,7 +43,7 @@
  * \include passthrough_hp.cc
  */
 
-#define FUSE_USE_VERSION 35
+#define FUSE_USE_VERSION FUSE_MAKE_VERSION(3, 12)
 
 #ifdef HAVE_CONFIG_H
 #include "config.h"
@@ -123,6 +123,8 @@ struct Inode {
     int fd {-1};
     dev_t src_dev {0};
     ino_t src_ino {0};
+    int generation {0};
+    uint64_t nopen {0};
     uint64_t nlookup {0};
     std::mutex m;
 
@@ -192,13 +194,19 @@ static void sfs_init(void *userdata, fuse_conn_info *conn) {
     if (conn->capable & FUSE_CAP_FLOCK_LOCKS)
         conn->want |= FUSE_CAP_FLOCK_LOCKS;
 
-    // Use splicing if supported. Since we are using writeback caching
-    // and readahead, individual requests should have a decent size so
-    // that splicing between fd's is well worth it.
-    if (conn->capable & FUSE_CAP_SPLICE_WRITE && !fs.nosplice)
-        conn->want |= FUSE_CAP_SPLICE_WRITE;
-    if (conn->capable & FUSE_CAP_SPLICE_READ && !fs.nosplice)
-        conn->want |= FUSE_CAP_SPLICE_READ;
+    if (fs.nosplice) {
+        // FUSE_CAP_SPLICE_READ is enabled in libfuse3 by default,
+        // see do_init() in in fuse_lowlevel.c
+        // Just unset both, in case FUSE_CAP_SPLICE_WRITE would also get enabled
+        // by detault.
+        conn->want &= ~FUSE_CAP_SPLICE_READ;
+        conn->want &= ~FUSE_CAP_SPLICE_WRITE;
+    } else {
+        if (conn->capable & FUSE_CAP_SPLICE_WRITE)
+            conn->want |= FUSE_CAP_SPLICE_WRITE;
+        if (conn->capable & FUSE_CAP_SPLICE_READ)
+            conn->want |= FUSE_CAP_SPLICE_READ;
+    }
 }
 
 
@@ -340,14 +348,29 @@ static int do_lookup(fuse_ino_t parent, const char *name,
     }
     e->ino = reinterpret_cast<fuse_ino_t>(inode_p);
     Inode& inode {*inode_p};
+    e->generation = inode.generation;
 
-    if(inode.fd != -1) { // found existing inode
+    if (inode.fd == -ENOENT) { // found unlinked inode
+        if (fs.debug)
+            cerr << "DEBUG: lookup(): inode " << e->attr.st_ino
+                 << " recycled; generation=" << inode.generation << endl;
+	/* fallthrough to new inode but keep existing inode.nlookup */
+    }
+
+    if (inode.fd > 0) { // found existing inode
         fs_lock.unlock();
         if (fs.debug)
             cerr << "DEBUG: lookup(): inode " << e->attr.st_ino
-                 << " (userspace) already known." << endl;
+                 << " (userspace) already known; fd = " << inode.fd << endl;
         lock_guard<mutex> g {inode.m};
+
         inode.nlookup++;
+        if (fs.debug)
+            cerr << "DEBUG:" << __func__ << ":" << __LINE__ << " "
+                 <<  "inode " << inode.src_ino
+                 << " count " << inode.nlookup << endl;
+
+
         close(newfd);
     } else { // no existing inode
         /* This is just here to make Helgrind happy. It violates the
@@ -357,13 +380,19 @@ static int do_lookup(fuse_ino_t parent, const char *name,
         lock_guard<mutex> g {inode.m};
         inode.src_ino = e->attr.st_ino;
         inode.src_dev = e->attr.st_dev;
-        inode.nlookup = 1;
+
+        inode.nlookup++;
+        if (fs.debug)
+            cerr << "DEBUG:" << __func__ << ":" << __LINE__ << " "
+                 <<  "inode " << inode.src_ino
+                 << " count " << inode.nlookup << endl;
+
         inode.fd = newfd;
         fs_lock.unlock();
 
         if (fs.debug)
             cerr << "DEBUG: lookup(): created userspace inode " << e->attr.st_ino
-                 << endl;
+                 << "; fd = " << inode.fd << endl;
     }
 
     return 0;
@@ -464,6 +493,10 @@ static void sfs_link(fuse_req_t req, fuse_ino_t ino, fuse_ino_t parent,
     {
         lock_guard<mutex> g {inode.m};
         inode.nlookup++;
+        if (fs.debug)
+            cerr << "DEBUG:" << __func__ << ":" << __LINE__ << " "
+                 <<  "inode " << inode.src_ino
+                 << " count " << inode.nlookup << endl;
     }
 
     fuse_reply_entry(req, &e);
@@ -496,6 +529,33 @@ static void sfs_rename(fuse_req_t req, fuse_ino_t parent, const char *name,
 
 static void sfs_unlink(fuse_req_t req, fuse_ino_t parent, const char *name) {
     Inode& inode_p = get_inode(parent);
+    // Release inode.fd before last unlink like nfsd EXPORT_OP_CLOSE_BEFORE_UNLINK
+    // to test reused inode numbers.
+    // Skip this when inode has an open file and when writeback cache is enabled.
+    if (!fs.timeout) {
+	    fuse_entry_param e;
+	    auto err = do_lookup(parent, name, &e);
+	    if (err) {
+		    fuse_reply_err(req, err);
+		    return;
+	    }
+	    if (e.attr.st_nlink == 1) {
+		    Inode& inode = get_inode(e.ino);
+		    lock_guard<mutex> g {inode.m};
+		    if (inode.fd > 0 && !inode.nopen) {
+			    if (fs.debug)
+				    cerr << "DEBUG: unlink: release inode " << e.attr.st_ino
+					    << "; fd=" << inode.fd << endl;
+			    lock_guard<mutex> g_fs {fs.mutex};
+			    close(inode.fd);
+			    inode.fd = -ENOENT;
+			    inode.generation++;
+		    }
+	    }
+
+        // decrease the ref which lookup above had increased
+        forget_one(e.ino, 1);
+    }
     auto res = unlinkat(inode_p.fd, name, 0);
     fuse_reply_err(req, res == -1 ? errno : 0);
 }
@@ -511,6 +571,12 @@ static void forget_one(fuse_ino_t ino, uint64_t n) {
         abort();
     }
     inode.nlookup -= n;
+
+    if (fs.debug)
+        cerr << "DEBUG:" << __func__ << ":" << __LINE__ << " "
+             <<  "inode " << inode.src_ino
+             << " count " << inode.nlookup << endl;
+
     if (!inode.nlookup) {
         if (fs.debug)
             cerr << "DEBUG: forget: cleaning up inode " << inode.src_ino << endl;
@@ -764,8 +830,13 @@ static void sfs_create(fuse_req_t req, fuse_ino_t parent, const char *name,
         if (err == ENFILE || err == EMFILE)
             cerr << "ERROR: Reached maximum number of file descriptors." << endl;
         fuse_reply_err(req, err);
-    } else
-        fuse_reply_create(req, &e, fi);
+	return;
+    }
+
+    Inode& inode = get_inode(e.ino);
+    lock_guard<mutex> g {inode.m};
+    inode.nopen++;
+    fuse_reply_create(req, &e, fi);
 }
 
 
@@ -814,14 +885,19 @@ static void sfs_open(fuse_req_t req, fuse_ino_t ino, fuse_file_info *fi) {
         return;
     }
 
+    lock_guard<mutex> g {inode.m};
+    inode.nopen++;
     fi->keep_cache = (fs.timeout != 0);
+    fi->noflush = (fs.timeout == 0 && (fi->flags & O_ACCMODE) == O_RDONLY);
     fi->fh = fd;
     fuse_reply_open(req, fi);
 }
 
 
 static void sfs_release(fuse_req_t req, fuse_ino_t ino, fuse_file_info *fi) {
-    (void) ino;
+    Inode& inode = get_inode(ino);
+    lock_guard<mutex> g {inode.m};
+    inode.nopen--;
     close(fi->fh);
     fuse_reply_err(req, 0);
 }
@@ -1152,6 +1228,8 @@ static void maximize_fd_limit() {
 
 int main(int argc, char *argv[]) {
 
+    struct fuse_loop_config *loop_config = NULL;
+
     // Parse command line options
     auto options {parse_options(argc, argv)};
 
@@ -1198,15 +1276,15 @@ int main(int argc, char *argv[]) {
     umask(0);
 
     // Mount and run main loop
-    struct fuse_loop_config loop_config;
-    loop_config.clone_fd = 0;
-    loop_config.max_idle_threads = 10;
+    loop_config = fuse_loop_cfg_create();
+
     if (fuse_session_mount(se, argv[2]) != 0)
         goto err_out3;
     if (options.count("single"))
         ret = fuse_session_loop(se);
     else
-        ret = fuse_session_loop_mt(se, &loop_config);
+        ret = fuse_session_loop_mt(se, loop_config);
+
 
     fuse_session_unmount(se);
 
@@ -1215,6 +1293,8 @@ err_out3:
 err_out2:
     fuse_session_destroy(se);
 err_out1:
+
+    fuse_loop_cfg_destroy(loop_config);
     fuse_opt_free_args(&args);
 
     return ret ? 1 : 0;
