@@ -351,10 +351,9 @@ int cuse_lowlevel_main(int argc, char *argv[], const struct cuse_info *ci,
 		return 1;
 
 	if (multithreaded) {
-		struct fuse_loop_config config;
-		config.clone_fd = 0;
-		config.max_idle_threads = 10;
-		res = fuse_session_loop_mt_32(se, &config);
+		struct fuse_loop_config *config = fuse_loop_cfg_create();
+		res = fuse_session_loop_mt(se, config);
+		fuse_loop_cfg_destroy(config);
 	}
 	else
 		res = fuse_session_loop(se);
